import numpy as np
from pyTurbulence.poissonSolver import poisson_3d, dilatational_velocities
from pyTurbulence.spectrum import energy_spectrum
from numba import jit, prange
from typing import Tuple

@jit(nopython=True, parallel=True)
def _compute_velocity_field(u_, v_, w_, nx, ny, nz, kx, ky, kz, 
                            xc, yc, zc, um, sxm, sym, szm, 
                            dx, dy, dz, psi):
    for k in prange(nz):  
        for j in prange(ny):
            for i in prange(nx):
                # Fourier summation
                arg = kx * xc[i] + ky * yc[j] + kz * zc[k] - psi
                bmx = 2.0 * um * np.cos(arg - kx * dx / 2.0)
                bmy = 2.0 * um * np.cos(arg - ky * dy / 2.0)
                bmz = 2.0 * um * np.cos(arg - kz * dz / 2.0)

                u_[i, j, k] = np.sum(bmx * sxm)
                v_[i, j, k] = np.sum(bmy * sym)
                w_[i, j, k] = np.sum(bmz * szm)

# MIT License
# Original function from turboGenPy by Tony Saad
# Modified by Amaury Bilocq in 2025
def compute_solenoidal_velocities(spectrum: str, nbModes: int, urms: float,
                                    k0: float, domain_size: Tuple[float, float, float], 
                                    domain_resolution: Tuple[int, int, int]
                                    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, float]:

    """
    Generate a synthetic turbulence velocity field using the Tony Saad's methodology.

    Parameters
    ----------
    spectrum : str
        Name of the energy spectrum to use.
    nbModes : int
        Number of modes to use in the synthetic turbulence generation.
    urms : float
        Root mean square velocity.
    k0 : float
        Peak wavenumber.
    domain_size : np.ndarray
        Physical size of the domain in each direction (Lx, Ly, Lz).
    domain_resolution : np.ndarray
        Number of grid points in each direction (nx, ny, nz).

    Returns
    -------
    u_ : np.ndarray
        Velocity component in the x-direction.
    v_ : np.ndarray
        Velocity component in the y-direction.
    w_ : np.ndarray
        Velocity component in the z-direction.
    wmax : float
        Maximum wavenumber.
    """

    Lx, Ly, Lz = domain_size
    nx, ny, nz = domain_resolution
    dx = Lx/nx
    dy = Ly/ny
    dz = Lz/nz

    # Proper wavenumber scaling
    # unique_kk, wave_instances, wave_width = wavespace(nbModes)
    wmin = 2.*np.pi/np.min([Lx,Ly,Lz])
    wmax = np.pi/np.min([dx,dy,dz])

    dk = (wmax - wmin)/nbModes
    wn = wmin + np.arange(0,nbModes)*dk
    dkn = np.ones(nbModes)*dk

    # Generate random angles - phases
    phi = 2.0 * np.pi * np.random.uniform(0.0, 1.0, nbModes)
    nu = np.random.uniform(0.0, 1.0, nbModes)
    theta = np.arccos(2.0 * nu - 1.0)
    psi = np.random.uniform(-np.pi / 2.0, np.pi / 2.0, nbModes)
    
    kx = np.sin(theta) * np.cos(phi) * wn
    ky = np.sin(theta) * np.sin(phi) * wn
    kz = np.cos(theta) * wn 

    # create divergence vector
    ktx = np.sin(kx * dx / 2.0) / dx
    kty = np.sin(ky * dy / 2.0) / dy
    ktz = np.sin(kz * dz / 2.0) / dz

    # Enforce Mass Conservation
    phi1 = 2.0 * np.pi * np.random.uniform(0.0, 1.0, nbModes)
    nu1 = np.random.uniform(0.0, 1.0, nbModes)
    theta1 = np.arccos(2.0 * nu1 - 1.0)
    zetax = np.sin(theta1) * np.cos(phi1)
    zetay = np.sin(theta1) * np.sin(phi1)
    zetaz = np.cos(theta1)
    sxm = zetay * ktz - zetaz * kty
    sym = -(zetax * ktz - zetaz * ktx)
    szm = zetax * kty - zetay * ktx
    smag = np.sqrt(sxm * sxm + sym * sym + szm * szm)
    sxm = sxm / smag
    sym = sym / smag
    szm = szm / smag

    # get the modes
    km = wn

    # compute the energy spectrum
    Ek = energy_spectrum(spectrum, km, urms, k0)
    Ek = Ek.clip(0.0)

    # generate turbulence
    um = np.sqrt(Ek * dkn)
    u_ = np.zeros((nx, ny, nz))
    v_ = np.zeros((nx, ny, nz))
    w_ = np.zeros((nx, ny, nz))

    # Generate coordinate grids (broadcasting-ready)
    xc = np.linspace(0, Lx, nx)
    yc = np.linspace(0, Ly, ny)
    zc = np.linspace(0, Lz, nz)

    # Parallelized loop using prange
    _compute_velocity_field(u_, v_, w_, nx, ny, nz, kx, ky, kz, xc, yc, zc, um, sxm, sym, szm, dx, dy, dz, psi)

    return u_, v_, w_,wmax

def compute_solenoidal_pressure(u: np.ndarray,v: np.ndarray,w: np.ndarray,
                                domain_resolution: Tuple[int,int,int],
                                domain_size: Tuple[float,float,float]) -> np.ndarray:
    """
    Compute the solenoidal pressure field from the velocity components.

    Parameters
    ----------
    u : np.ndarray
        Velocity component in the x-direction.
    v : np.ndarray
        Velocity component in the y-direction.
    w : np.ndarray
        Velocity component in the z-direction.
    domain_resolution : tuple
        Number of grid points in each direction (nx, ny, nz).
    domain_size : tuple
        Physical size of the domain in each direction (Lx, Ly, Lz).

    Returns
    -------
    pressure : np.ndarray
        Solenoidal pressure field.
    """

    nx = domain_resolution[0]
    ny = domain_resolution[1]
    nz = domain_resolution[2]

    hx = domain_size[0]/nx
    hy = domain_size[1]/ny
    hz = domain_size[2]/nz

    dudx,dudy,dudz = np.gradient(u,hx,hy,hz,edge_order=2)
    dvdx,dvdy,dvdz = np.gradient(v,hx,hy,hz,edge_order=2)
    dwdx,dwdy,dwdz = np.gradient(w,hx,hy,hz,edge_order=2)

    grad = np.array([[dudx,dudy,dudz],[dvdx,dvdy,dvdz],[dwdx,dwdy,dwdz]])

    rhs = np.zeros([nx,ny,nz])
    for i in range(3):
        for j in range(3):
            rhs += -1.*grad[i][j]*grad[j][i]

    pressure = poisson_3d(rhs, nx, ny, nz, hx, hy, hz)
    return pressure

def compute_thermodynamic_fields(mean_density : float, mean_pressure : float, mean_temperature : float,
                                 incompressible_pressure_fluctuations: np.ndarray, 
                                 gamma: float, Mt : float, case: int)->Tuple[np.ndarray,np.ndarray,np.ndarray]:
    """
    Compute the density, pressure and temperature fields from the incompressible pressure fluctuations.
    Eq (4-5) + ansatz from Ristorcelli and Blaisdell : p' = gamma * Mt^2 * p1
    
    Parameters
    ----------
    mean_density : float
        Mean density.
    mean_pressure : float
        Mean pressure.
    mean_temperature : float
        Mean temperature.
    incompressible_pressure_fluctuations : np.ndarray
        Incompressible pressure fluctuations.
    gamma : float
        Heat capacity ratio.
    Mt : float
        Mach number.
    case : int
        Case number.  

    Returns
    -------
    density : np.ndarray
        Density field.
    pressure : np.ndarray
        Pressure field.
    temperature : np.ndarray
        Temperature field.
    """
    compressible_pressure_fluctuations = incompressible_pressure_fluctuations * gamma * Mt**2 

    if case == 1:
        compressible_density_fluctuations = compressible_pressure_fluctuations
        compressible_temperature_fluctuations = compressible_pressure_fluctuations

    if case == 2:
        compressible_density_fluctuations = compressible_pressure_fluctuations*Mt**2
        compressible_temperature_fluctuations = compressible_pressure_fluctuations*Mt**2*(gamma-1)

    # Get fields of density - temperature - pressure
    density     = mean_density*(1.+compressible_density_fluctuations)
    temperature = mean_temperature*(1.+compressible_temperature_fluctuations)
    pressure    = mean_pressure*(1.+compressible_pressure_fluctuations)

    return density, pressure, temperature

def compute_dilatational_velocities(u: np.ndarray,v: np.ndarray,w: np.ndarray,p: np.ndarray,
                                    gamma: float, domain_resolution: Tuple[int,int,int],
                                    domain_size: Tuple[float,float,float]) -> np.ndarray:
    """
    Compute the dilatational velocity field from the velocity components and the pressure field.
    First, Eq.7 from Ristorcelli and Blaisdell is solved to get the dilatation field.
    Then, the dilatational velocity field is computed using the dilatation field (Eq.12).

    Parameters
    ----------
    u : np.ndarray
        Velocity component in the x-direction.
    v : np.ndarray
        Velocity component in the y-direction.
    w : np.ndarray
        Velocity component in the z-direction.
    p : np.ndarray
        Pressure field.
    gamma : float
        Heat capacity ratio.
    domain_resolution : tuple
        Number of grid points in each direction (nx, ny, nz).
    domain_size : tuple
        Physical size of the domain in each direction (Lx, Ly, Lz).

    Returns
    -------
    dilatational_velocity : np.ndarray
        Dilatational velocity field.
    """

    nx = domain_resolution[0]
    ny = domain_resolution[1]
    nz = domain_resolution[2]

    hx = domain_size[0]/nx
    hy = domain_size[1]/ny
    hz = domain_size[2]/nz

    overGamma = 1./gamma

    velocity       = np.array([u,v,w])
    gradVelocity   = np.array([np.gradient(vel, hx, hy, hz, edge_order=2) for vel in velocity])
    gradPressure   = np.gradient(p,hx,hy,hz,edge_order=2)

    # Compute time derivative of pressure fluctuations (Eq.11)
    #rhs = ((u_k * du_i/dx_k + dp/dx_i) * u_j)
    rhs = np.einsum('k...,ik...->i...', velocity, gradVelocity) + gradPressure
    rhs = np.einsum('i...,j...->...', rhs, velocity)

    # compute second derivative of rhs 
    # (p,t),jj = 2.*(rhs),ij
    grad_rhs    = np.gradient(rhs, hx, hy, hz, edge_order=2)
    hessian_rhs = np.array([np.gradient(grad, hx, hy, hz, edge_order=2) for grad in grad_rhs])
    rhs_ij      = np.einsum('ij...->...', hessian_rhs) 
    rhs_ij      *= 2.0
    ddt_p = poisson_3d(rhs_ij, nx, ny, nz, hx, hy, hz)

    # Compute dilatation from Eq.7
    # -gamma*dilatation = ddt_p+v_k*dpdxk
    dilatation = -overGamma*(np.sum(velocity * gradPressure, axis=0) + ddt_p)

    # Compute dilatational velocity from Eq.12
    vd_x, vd_y, vd_z = dilatational_velocities(dilatation, nx, ny, nz, hx, hy, hz)
    return vd_x, vd_y, vd_z
